(*== Church encodings in Elementary Linear Logic ==*)

(* v.0.3 Olivier Laurent *)

Require Import stdlib_ell2coq. (* v.0.5.6-1 *)


(*= Booleans =*)

Definition CBool := forall X : Prop, !X -> !X -> !X.

Fact ctrue : CBool.
Proof.
intros X x y.
lin_weak y.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact cfalse : CBool.
Proof.
intros X x y.
lin_weak x.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact cif : forall a : Prop, CBool -> !a -> !a -> !a.
Proof.
intros a b x y.
lin_apply2 b with x ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact cnot : CBool -> CBool.
Proof.
intro b.
intros X x y.
lem_apply3 (cif X) with b :: x ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact cor : CBool -> CBool -> CBool.
Proof.
intros b1 b2.
intros X x y.
lin_contr x.
lem_apply3 (cif X) with b1 :: x0.
lin_assumption.
lin_assumption.
lem_apply3 (cif X) with b2 :: x ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact cand : CBool -> CBool -> CBool.
Proof.
intros b1 b2.
intros X x y.
lin_contr y.
lem_apply3 (cif X) with b1 :: b2 x y.
lin_assumption.
lem_apply3 (cif X) with b2 :: x ;
lin_assumption.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.


(*= Unary natural numbers =*)

Definition CNat := forall X : Prop, !(X -> X) -> !(X -> X).

Fact czero : CNat.
Proof.
intros X f.
lin_weak f.
lin_mfprom_oc.
intro x.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact cone : CNat.
Proof.
intros X f.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact ctwo : CNat.
Proof.
intros X f.
lin_contr f.
lin_mfprom_oc.
intro x.
lin_apply1 f0.
lin_apply1 f.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact cmult : CNat -> CNat -> CNat.
Proof.
intros m n.
intros X f.
(* llcheck
specialize (m X).
specialize (n X).
llcheck *)
lin_apply1 m.
lin_apply1 n.
lin_assumption.
(* llcheck
ELL.
llcheck *)
(* BUG: specialize required *)
Qed.
(*
Defined.
Eval compute in cmult.
Print cmult.
*)

Fact cadd : CNat -> CNat -> CNat.
Proof.
intros m n.
intros X f.
lin_contr f.
lin_hyp_apply1 m in f.
lin_hyp_apply1 n in f0.
lin_mfprom_oc.
intro x.
lin_apply1 f0.
lin_apply1 f.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.
(*
Defined.
Eval compute in cadd.
*)

Fact csucc : CNat -> CNat.
Proof.
intro n.
lem_apply2 cadd with.
lem_apply0 cone.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Ltac iter n :=
  lazymatch goal with
  | |- OC ?A -> OC ?A => specialize (n A) ;
                         lem_apply1 mon_oc_lmap ;
                         lin_apply1 n ;
                         lin_mfprom_oc
  | _                 => fail "Conclusion is not of the shape !A -> !A"
  end.

Fact cnat_coer : CNat -> !CNat.
Proof.
intro n.
lin_cut (!CNat) as i with n.
revert i.
iter n.
lem_apply0 csucc.
lin_mfprom_oc.
lem_apply0 czero.
(* llcheck
ELL.
llcheck *)
Qed.


(* Polynomials *)

Fact oc_cadd : !CNat -> !CNat -> !CNat.
Proof.
intros n m.
lin_mfprom_oc.
lem_apply2 cadd with n ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact oc_cmult : !CNat -> !CNat -> !CNat.
Proof.
intros n m.
lin_mfprom_oc.
lem_apply2 cmult with n ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Ltac nat2cnat n :=
  match n with
  | 0    => lem_apply0 czero
  | S ?k => lem_apply2 cadd with ;
            [ lem_apply0 cone | nat2cnat k]
  end.

Ltac cpower n :=
  let HX := fresh "X" in
  lin_cut (!CNat) as HX with ;
  [ match n with
    | 0    => lin_weak HX ;
              lin_mfprom ;
              lem_apply0 cone
    | S ?k => lin_contr HX ;
              lem_apply2 oc_cmult with HX ;
              [ lin_assumption | cpower k ; lin_assumption ]
    end
  | idtac ].

Ltac cmonom p :=
  match p with
  | (?n,?m) => lem_apply2 oc_cmult with ;
               [ lin_mfprom_oc ; nat2cnat n
               | cpower m ]
  end.

Ltac cpoly l :=
  let HX := fresh "X" in
  lin_cut (!CNat) as HX with ;
  [ match l with
    | nil        => lin_weak HX ;
                    lin_mfprom_oc ;
                    lem_apply0 czero
    | cons ?p ?t => lin_contr HX ;
                    lem_apply2 oc_cadd with HX ;
                    [ cmonom p ; lin_assumption
                    | cpoly t ; lin_assumption ]
    end
  | idtac ].


(*= Binary natural numbers =*)

Definition CBin := forall X : Prop, !(X -> X) -> !(X -> X) -> !(X -> X).

Fact cnil : CBin.
Proof.
intros X c0 c1.
lin_weak c0.
lin_weak c1.
lin_mfprom_oc.
intro n.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact ccons0 : CBin -> CBin.
Proof.
intros l X c0 c1.
lin_contr c0.
specialize (l X).
lin_left_imp l with c1 c2.
lin_left_imp l with c2.
lin_mfprom_oc.
intro n.
lin_apply1 l.
lin_apply1 c2.
lin_assumption.
lin_assumption.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact ccons1 : CBin -> CBin.
Proof.
intros l X c0 c1.
lin_contr c1.
specialize (l X).
lin_left_imp l with c1 c2.
lin_left_imp l with c2.
lin_mfprom_oc.
intro n.
lin_apply1 l.
lin_apply1 c2.
lin_assumption.
lin_assumption.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact cappend : CBin -> CBin -> CBin.
Proof.
intros l1 l2.
intros X c0 c1.
lin_contr c0.
lin_contr c1.
specialize (l1 X).
specialize (l2 X).
lin_left_imp l1 with l2 c1 c2 c3.
lin_left_imp l1 with l2 c2 c3.
lin_left_imp l2 with l1 c2.
lin_left_imp l2 with l1.
lin_mfprom_oc.
intro x.
lin_apply1 l2.
lin_apply1 l1.
lin_assumption.
lin_assumption.
lin_assumption.
lin_assumption.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact clength : CBin -> CNat.
Proof.
intro l.
intros X s.
lin_contr s.
lin_apply2 l with s ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Ltac iter2 l :=
  lazymatch goal with
  | |- OC ?A -> OC ?A => specialize (l A) ;
                         lem_apply1 mon_oc_lmap ;
                         (lin_apply2 l with ; lin_mfprom_oc)
  | _                 => fail "Conclusion is not of the shape !A -> !A"
  end.

Fact cbin_coer : CBin -> !CBin.
Proof.
intros l.
lin_cut (!CBin) as i with l.
revert i.
iter2 l.
lem_apply0 ccons0.
lem_apply0 ccons1.
lin_mfprom_oc.
lem_apply0 cnil.
(* llcheck
ELL.
llcheck *)
Qed.


