(*== Kalmar functions in Elementary Linear Logic ==*)
(*== with first-order constaints in the style of AF2 ==*)

(*= inspired by  http://dx.doi.org/10.1007/978-3-642-21691-6_12  =*)
(*=          and http://dx.doi.org/10.1016/S0890-5401(03)00010-5 =*)

(* v.0.1 Olivier Laurent *)

Require Import stdlib_ell2coq. (* v.0.5.6-1 *)

Parameter N : Type.
Parameter N0 : N.
Parameter NS : N -> N.

Notation N1 := (NS N0).
Notation N2 := (NS N1).

Definition CNat (x : N) :=
  forall X : N -> Prop, !(forall y : N, (X y -> X (NS y))) -> !(X N0 -> X x).

Fact czero : CNat N0.
Proof.
intros X f.
lin_weak f.
lin_mfprom_oc.
intro x.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact cone : CNat N1.
Proof.
intros X f.
lin_mfprom_oc.
intro x.
lin_apply1 f.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact ctwo : CNat N2.
Proof.
intros X f.
lin_contr f.
lin_mfprom_oc.
intro x.
lin_apply1 f0.
lin_apply1 f.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact csucc : forall a, CNat a -> CNat (NS a).
Proof.
intros a n.
intros X HS.
specialize (n (fun z => X (NS z))).
simpl in n.
lin_contr HS.
lin_cut (!(X N1 -> X (NS a))) as n' with HS0.
lin_mfprom_oc.
intro H0.
lin_apply1 n'.
lin_apply1 HS0.
lin_assumption.
lin_apply1 n.
lin_mfprom_oc.
intros y HSy.
lin_apply1 HS.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Parameter Npred : N -> N.
Axiom Npred_0 : Npred N0 = N0.
Axiom Npred_S : forall a, Npred (NS a) = a.

Fact cpred : forall a, CNat a -> CNat (Npred a).
Proof.
intros a n X HS.
specialize (n (fun z => (X (Npred z) [&] X z))).
simpl in n.
rewrite Npred_0 in n.
lin_cut (!(X N0 [&] X N0 -> X (Npred a) [&] X a)) as n' with.
lin_mfprom_oc.
intro H0.
lin_cut (X (Npred a) [&] X a) as p with.
lin_left_with_1 p.
lin_assumption.
lin_apply1 n'.
lin_with ;
lin_assumption.
lin_apply1 n.
lin_mfprom_sh.
intros y Hp.
lin_left_with_2 Hp.
lin_with.
lin_weak_sh HS.
rewrite Npred_S.
lin_assumption.
lin_der_sh HS.
lin_apply1 HS.
lin_assumption.
(* llcheck
EsLL.
llcheck *)
Qed.

Fact cpred_aff : forall a, CNat a -> CNat (Npred a).
(* affine proof *)
Proof.
intros a n X HS.
specialize (n (fun z => ((X (Npred z) -> X z) /\ X (Npred z)))).
simpl in n.
rewrite Npred_0 in n.
lin_cut (!((X N0 -> X N0) /\ X N0 -> (X (Npred a) -> X a) /\ X (Npred a))) as n' with.
lin_mfprom_oc.
intro H0.
lin_cut ((X (Npred a) -> X a) /\ X (Npred a)) as p with.
lin_left_conj p.
clear p0. (* affine *)
lin_assumption.
lin_apply1 n'.
lin_split with.
intro.
lin_assumption.
lin_assumption.
lin_apply1 n.
lin_mfprom_oc.
intros y Hp.
lin_left_conj Hp.
rewrite Npred_S.
lin_split with HS.
lin_apply0 HS.
lin_apply1 Hp0.
lin_assumption.
(* llcheck
EAL.
llcheck *)
Qed.

Parameter Nadd : N -> N -> N.
Axiom Nadd_0 : forall a, Nadd N0 a = a.
Axiom Nadd_S : forall a b, Nadd (NS a) b = NS (Nadd a b).

Fact Nadd_1 : forall a, Nadd N1 a = NS a.
Proof.
intro a.
rewrite Nadd_S.
rewrite Nadd_0.
reflexivity.
(* llcheck
ELL.
llcheck *)
Qed.

Fact cadd : forall a b, CNat a -> CNat b -> CNat (Nadd a b).
Proof.
intros a b m n.
intros X f.
lin_contr f.
specialize (m (fun z => X (Nadd z b))).
simpl in m.
lin_cut (!(X (Nadd N0 b) -> X (Nadd a b))) as m' with n f0.
lin_hyp_apply1 n in f0.
lin_mfprom_oc.
intro x.
lin_apply1 m'.
rewrite Nadd_0.
lin_apply1 f0.
lin_assumption.
lin_apply1 m.
lin_mfprom_oc.
intros y H.
rewrite Nadd_S.
lin_apply1 f.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact csucc_add : forall a, CNat a -> CNat (NS a).
Proof.
intros a n.
rewrite <- Nadd_1.
lem_apply2 cadd with.
lem_apply0 cone.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Parameter Nmult : N -> N -> N.
Axiom Nmult_0 : forall a, Nmult N0 a = N0.
Axiom Nmult_S : forall a b, Nmult (NS a) b = Nadd b (Nmult a b).

(* intuitionistic proof non compatible with ELL
   while the propositional version is *)
(*
Definition Nat (x : N) :=
  forall X : N -> Prop, (forall y : N, (X y -> X (NS y))) -> (X N0 -> X x).

Fact cmult1 : forall a b, Nat a -> Nat b -> Nat (Nmult a b).
Proof.
intros a b m n.
intros X f.
specialize (m (fun z => X (Nmult z b))) ; simpl in m ;
  rewrite Nmult_0 in m.
apply m ; clear m.
intro y. (* mismatch here between introduction of y and of ! *)
rewrite Nmult_S.
specialize (n (fun z => X (Nadd z (Nmult y b)))) ; simpl in n ;
  rewrite Nadd_0 in n.
apply n ; clear n.
intros y0.
rewrite Nadd_S.
apply f.
Qed.
*)

Fact cmult : forall a b , CNat a -> CNat b -> CNat (Nmult a b).
Proof.
intros a b m n.
intros X f.
specialize (m (fun z => X (Nmult z b))).
simpl in m.
rewrite Nmult_0 in m.
lin_apply1 m.
specialize (n (fun z => forall y : N, X y -> X (Nadd z y))).
simpl in n.
lin_cut (!((forall y : N, X y -> X (Nadd N0 y)) ->
      forall y : N, X y -> X (Nadd b y))) as n' with.
lin_mfprom_oc.
intros y H.
rewrite Nmult_S.
lin_apply2 n' with.
intros y H.
rewrite Nadd_0.
lin_assumption.
lin_assumption.
lin_apply1 n.
lin_mfprom_oc.
intros y g y0 z.
rewrite Nadd_S.
lin_apply1 f.
lin_apply1 g.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Ltac iter n :=
  lazymatch goal with
  | |- OC (?A _) -> OC (?A _) => 
           let x := fresh "x" in
             specialize (n A) ;
             lem_apply1 mon_oc_lmap ;
             lin_apply1 n ;
             lin_mfprom_oc ;
             intro x
  | _   => fail "Conclusion is not of the shape !(A _) -> !(A _)"
  end.

Ltac iter_spec n A :=
  lazymatch goal with
  | |- OC _ -> OC _ => 
           let x := fresh "x" in
             specialize (n A) ;
             lem_apply1 mon_oc_lmap ;
             lin_apply1 n ;
             lin_mfprom_oc ;
             intro x
  | _   => fail "Conclusion is not of the shape !(A _) -> !(A _)"
  end.

Fact coer : forall a , CNat a -> !CNat a.
Proof.
intros a n.
lin_cut (!(CNat N0)) as i with n.
revert i.
iter n.
lem_apply0 csucc.
lin_mfprom_oc.
lem_apply0 czero.
(* llcheck
ELL.
llcheck *)
Qed.

Parameter Nminus : N -> N -> N.
Axiom Nminus_0  : forall a, Nminus a N0 = a.
Axiom Nminus_S  : forall a b, Nminus a (NS b) = Npred (Nminus a b).

Fact cminus : forall a b : N, CNat a -> CNat b -> !CNat (Nminus a b).
Proof.
intros a b m n.
lin_cut (!CNat (Nminus a N0)) as m' with n.
revert m'.
iter_spec n (fun x => CNat (Nminus a x)).
intro H.
rewrite Nminus_S.
lem_apply1 cpred.
lin_assumption.
lem_apply1 coer.
rewrite Nminus_0.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact cproj_one : forall a b, CNat a -> CNat b -> !CNat a.
Proof.
intros a b m n.
lin_cut (!CNat a) as m' with n.
lin_cut (!CNat b) as n' with m'.
lin_weak n'.
lin_assumption.
lem_apply1 coer.
lin_assumption.
lem_apply1 coer.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact cproj_two : forall a b, CNat a -> CNat b -> !CNat b.
Proof.
intros a b m n.
lin_cut (!CNat a) as m' with n.
lin_cut (!CNat b) as n' with m'.
lin_weak m'.
lin_assumption.
lem_apply1 coer.
lin_assumption.
lem_apply1 coer.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Ltac coer_it :=
  lazymatch goal with
  | |- !^(?n) (CNat ?a) => 
    match n with
    | 0    => simpl
    | S ?k => let Hn := fresh in
              lin_cut (!^(k)(CNat a)) as Hn with ;
              [ simpl ;
                simpl in Hn ;
                repeat lin_mfprom_oc ;
                lem_apply1 coer ;
                lin_assumption
              | coer_it ]
    end
  | _ => fail "Goal not of the shape !^(n)(CNat _)"
  end.

Ltac comp_unary :=
  let Hf := fresh in
  let Ha := fresh in
  let H := fresh in
  lazymatch goal with
  | |- !^(?m)!^(?n) (CNat (?f ?a))
      => lin_cut ((!^(m)CNat a) -> !^(m)!^(n)(CNat (f a))) as Hf with ;
         [ lin_apply1 Hf
         | simpl ;
           intro Ha ;
           repeat lin_mfprom_oc ;
           lin_cut (!^(n)(CNat (f a))) as H with ;
           [ simpl in H ; lin_assumption
           | idtac ]
         ]
  end.

Ltac comp_binary :=
  let Hf := fresh in
  let Ha := fresh in
  let Hb := fresh in
  let H := fresh in
  lazymatch goal with
  | |- !^(?m)!^(?n) (CNat (?f ?a ?b))
      => lin_cut
           ((!^(m)CNat a) -> (!^(m)CNat b) -> !^(m)!^(n)(CNat (f a b)))
           as Hf with ;
         [ lin_apply2 Hf with
         | simpl ;
           intros Ha Hb ;
           repeat lin_mfprom_oc ;
           lin_cut (!^(n)(CNat (f a b))) as H with ;
           [ simpl in H ; lin_assumption
           | idtac ]
         ]
  end.

Parameter Nsum : (N -> N) -> N -> N.
Axiom Nsum_0 : forall f, Nsum f N0 = N0.
Axiom Nsum_S : forall f a, Nsum f (NS a) = Nadd (Nsum f a) (f a).

Ltac csum n :=
(* turn goal  n:CNat a |- !!!^(k)(CNat (Nsum f a))
   with a concrete k into  |- forall x, CNat x -> !^(k)CNat (f x)*)
lazymatch goal with
| |- !!!^(?k)(CNat (Nsum ?f ?a)) =>
let p := fresh in
let HI := fresh in
let Hf := fresh in
let y := fresh in
let Hy := fresh in
let Hy0 := fresh in
lin_cut (!!forall x, CNat x -> !^(k)CNat (f x)) as Hf with n ;
[ lin_cut (!(!((CNat N0 [&] !^(k)CNat (Nsum f N0)))
             -> !(CNat a [&] !^(k)CNat (Nsum f a)))) as HI with ;
  [ lin_mfprom_oc ;
    lin_cut (!(CNat a [&] !^(k)CNat (Nsum f a))) as p with ;
    [ lin_mfprom_oc ;
      lin_left_with_2 p ;
      lin_assumption
    | lin_apply1 HI ;
      lin_mfprom_oc ;
      lin_with ;
      [ lem_apply0 czero
      | rewrite Nsum_0 ;
        repeat lin_mfprom_oc ;
        lem_apply0 czero ]
    ]
  | specialize (n (fun z => (!(CNat z [&] !^(k)CNat (Nsum f z))))) ;
    lin_left_imp n with ;
    [ lin_assumption
    | lin_mfprom_oc ;
      intros y Hy ;
      lin_mfprom_sh ;
      lin_with ;
      [ lin_weak_sh Hf ;
        lin_der_sh Hy ;
        lin_left_with_1 Hy ;
        lem_apply1 csucc ;
        lin_assumption
      | rewrite Nsum_S ;
        lin_der_sh Hf ;
        lin_contr_sh Hy ;
        lin_cut (#(CNat y [&] !^(k)CNat (Nsum f y))) as Hy0 with Hf Hy ;
        [ lin_der_sh Hy ;
          lin_der_sh Hy0 ;
          lin_left_with_1 Hy ;
          lin_left_with_2 Hy0 ;
          lin_hyp_apply1 Hf in Hy ;
          simpl in Hy ;
          simpl in Hy0 ;
          repeat lin_mfprom_oc ;
          lem_apply2 cadd with Hy0 ;
          lin_assumption
        | lin_assumption ]
      ]
    ]
  ]
| lin_mfprom_oc ;
  lin_mfprom_oc ]
end.

Parameter Nprod : (N -> N) -> N -> N.
Axiom Nprod_0 : forall f, Nprod f N0 = N1.
Axiom Nprod_S : forall f a, Nprod f (NS a) = Nmult (Nprod f a) (f a).

Ltac cprod n :=
(* turn goal  n:CNat a |- !!!^(k)(CNat (Nprod f a))
   with a concrete k into  |- forall x, CNat x -> !^(k)CNat (f x)*)
lazymatch goal with
| |- !!!^(?k)(CNat (Nprod ?f ?a)) =>
let p := fresh in
let HI := fresh in
let Hf := fresh in
let y := fresh in
let Hy := fresh in
let Hy0 := fresh in
lin_cut (!!forall x, CNat x -> !^(k)CNat (f x)) as Hf with n ;
[ lin_cut (!(!((CNat N0 [&] !^(k)CNat (Nprod f N0)))
             -> !(CNat a [&] !^(k)CNat (Nprod f a)))) as HI with ;
  [ lin_mfprom_oc ;
    lin_cut (!(CNat a [&] !^(k)CNat (Nprod f a))) as p with ;
    [ lin_mfprom_oc ;
      lin_left_with_2 p ;
      lin_assumption
    | lin_apply1 HI ;
      lin_mfprom_oc ;
      lin_with ;
      [ lem_apply0 czero
      | rewrite Nprod_0 ;
        repeat lin_mfprom_oc ;
        lem_apply0 cone ]
    ]
  | specialize (n (fun z => (!(CNat z [&] !^(k)CNat (Nprod f z))))) ;
    lin_left_imp n with ;
    [ lin_assumption
    | lin_mfprom_oc ;
      intros y Hy ;
      lin_mfprom_sh ;
      lin_with ;
      [ lin_weak_sh Hf ;
        lin_der_sh Hy ;
        lin_left_with_1 Hy ;
        lem_apply1 csucc ;
        lin_assumption
      | rewrite Nprod_S ;
        lin_der_sh Hf ;
        lin_contr_sh Hy ;
        lin_cut (#(CNat y [&] !^(k)CNat (Nprod f y))) as Hy0 with Hf Hy ;
        [ lin_der_sh Hy ;
          lin_der_sh Hy0 ;
          lin_left_with_1 Hy ;
          lin_left_with_2 Hy0 ;
          lin_hyp_apply1 Hf in Hy ;
          simpl in Hy ;
          simpl in Hy0 ;
          repeat lin_mfprom_oc ;
          lem_apply2 cmult with Hy0 ;
          lin_assumption
        | lin_assumption ]
      ]
    ]
  ]
| lin_mfprom_oc ;
  lin_mfprom_oc ]
end.


