(*=== Automation tactics for l2coq ===*)

(* v.0.4 Olivier Laurent *)

Require Export l2coq. (* v.0.5.6 *)


(*= MALL =*)

Ltac lin_easy_mall := (try lin_top) ;
                      (try lin_one) ;
                      (try lin_assumption).

Ltac lin_rever_mall :=
  intros ;
  match goal with
  | |- _ [&] _ => lin_with ; lin_rever_mall
  | |- [T] => lin_top
  | H : _ /\ _ |- _ => lin_left_conj H ; lin_rever_mall
  | H : True |- _ => lin_left_one H ; lin_rever_mall
  | H : _ [+] _ |- _ => lin_left_plus H ; lin_rever_mall
  | H : [0] |- _ => lin_left_zero H
  | H : exists _, _ |- _ => destruct H ; lin_rever_mall
  | |- _ <-> _ => no_hyps ; lin_split with ; lin_rever_mall
  | |- addiff _ _ => unfold addiff ; lin_rever_mall
  | |- iff _ _ => unfold iff ; lin_rever_mall
  | |- _ /\ _ => no_hyps ; lin_split with ; lin_rever_mall
  | |- True => no_hyps ; lin_one
  | _ => try lin_assumption
  end.

Ltac lin_auto_mall :=
  lin_rever_mall ;
  match goal with
  | |- _ <-> _ => try (lin_split with ; lin_auto_mall)
  | |- _ /\ _ => try (lin_split with ; lin_auto_mall)
  | |- _ [+] _ => try (lin_plus_1 ; lin_auto_mall) ;
                  try (lin_plus_2 ; lin_auto_mall)
  | H : _ [&] _ |- _ => try (lin_left_with_1 H ; lin_auto_mall) ;
                        try (lin_left_with_2 H ; lin_auto_mall)
  | H : _ -> _ |- _ => try (lin_left_imp H with ; lin_auto_mall)
  end ;
  fail "Goal not solved"
.


(*= LLL =*)

Ltac lin_easy_lll := (try lin_top) ;
                     (try lin_one_oc) ;
                     (try lin_assumption_oc).

Ltac lin_rever_lll :=
  intros ;
  match goal with
  | |- _ [&] _ => lin_with ; lin_rever_lll
  | |- [T] => lin_top
  | H : _ /\ _ |- _ => lin_left_conj H ; lin_rever_lll
  | H : True |- _ => lin_left_one H ; lin_rever_lll
  | H : _ [+] _ |- _ => lin_left_plus H ; lin_rever_lll
  | H : [0] |- _ => lin_left_zero H
  | H : exists _, _ |- _ => destruct H ; lin_rever_lll
  | |- PG _ => are_pg_hyps ; lin_mfprom_pg ; lin_rever_lll
  | |- OC _ => no_hyps ; lin_ufprom_oc ; lin_rever_lll
  | |- _ <-> _ => no_hyps ; lin_split with ; lin_rever_lll
  | |- addiff _ _ => unfold addiff ; lin_rever_lll
  | |- iff _ _ => unfold iff ; lin_rever_lll
  | |- _ /\ _ => no_hyps ; lin_split with ; lin_rever_lll
  | |- True => no_hyps ; lin_one
  | _ => lin_easy_lll
  end.

Ltac lin_auto_lll :=
  lin_rever_lll ;
  match goal with
  | |- _ <-> _ => try (lin_split with ; lin_auto_lll)
  | |- _ /\ _ => try (lin_split with ; lin_auto_lll)
  | |- _ [+] _ => try (lin_plus_1 ; lin_auto_lll) ;
                  try (lin_plus_2 ; lin_auto_lll)
  | H : _ [&] _ |- _ => try (lin_left_with_1 H ; lin_auto_lll) ;
                        try (lin_left_with_2 H ; lin_auto_lll)
  | H : _ -> _ |- _ => try (lin_left_imp H with ; lin_auto_lll)
  | |- PG _ => try (lin_mfprom_pg ; lin_auto_lll)
  | |- OC _ => try (lin_ufprom_oc ; lin_auto_lll)
  end ;
  fail "Goal not solved"
.


(*= ELL =*)

Ltac lin_easy_sh := (try lin_top) ;
                    (try lin_one_sh) ;
                    (try lin_assumption_sh) ;
                    (try lin_assumption_der_sh).

Ltac lin_rever_ell :=
  intros ;
  match goal with
  | |- _ [&] _ => lin_with ; lin_rever_ell
  | |- [T] => lin_top
  | H : _ /\ _ |- _ => lin_left_conj H ; lin_rever_ell
  | H : True |- _ => lin_left_one H ; lin_rever_ell
  | H : _ [+] _ |- _ => lin_left_plus H ; lin_rever_ell
  | H : [0] |- _ => lin_left_zero H
  | H : exists _, _ |- _ => destruct H ; lin_rever_ell
  | |- PG _ => are_pg_hyps ; lin_mfprom_pg ; lin_rever_ell
  | |- OC _ => no_hyps ; lin_ufprom_oc ; lin_rever_lll
  | |- _ <-> _ => no_hyps ; lin_split with ; lin_rever_ell
  | |- addiff _ _ => unfold addiff ; lin_rever_ell
  | |- iff _ _ => unfold iff ; lin_rever_ell
  | |- _ /\ _ => no_hyps ; lin_split with ; lin_rever_ell
  | |- True => no_hyps ; lin_one
  | _ => lin_easy_sh
  end.

Ltac lin_auto_ell :=
  lin_rever_ell ;
  match goal with
  | |- _ <-> _ => try (lin_split with ; lin_auto_ell)
  | |- _ /\ _ => try (lin_split with ; lin_auto_ell)
  | |- _ [+] _ => try (lin_plus_1 ; lin_auto_ell) ;
                  try (lin_plus_2 ; lin_auto_ell)
  | H : _ [&] _ |- _ => try (lin_left_with_1 H ; lin_auto_ell) ;
                        try (lin_left_with_2 H ; lin_auto_ell)
  | H : _ -> _ |- _ => try (lin_left_imp H with ; lin_auto_ell)
  | |- PG _ => try (lin_mfprom_pg ; lin_auto_ell)
  | |- OC _ => try (lin_mfprom_oc ; lin_auto_ell) ;
               try (lin_mfprom_sh ; lin_auto_ell)
  end ;
  fail "Goal not solved"
.


(*= LL =*)

Ltac lin_easy := (try lin_top) ;
                 (try lin_one_oc) ;
                 (try lin_assumption_oc) ;
                 (try lin_assumption_der).

Ltac lin_rever :=
  intros ;
  match goal with
  | |- _ [&] _ => lin_with ; lin_rever
  | |- [T] => lin_top
  | H : _ /\ _ |- _ => lin_left_conj H ; lin_rever
  | H : True |- _ => lin_left_one H ; lin_rever
  | H : _ [+] _ |- _ => lin_left_plus H ; lin_rever
  | H : [0] |- _ => lin_left_zero H
  | H : exists _, _ |- _ => destruct H ; lin_rever
  | |- OC _ => lin_prom ; lin_rever
  | |- _ <-> _ => no_hyps ; lin_split with ; lin_rever
  | |- addiff _ _ => unfold addiff ; lin_rever
  | |- iff _ _ => unfold iff ; lin_rever
  | |- _ /\ _ => no_hyps ; lin_split with ; lin_rever
  | |- True => no_hyps ; lin_one
  | _ => lin_easy
  end.

Ltac lin_auto :=
  lin_rever ;
  match goal with
  | |- _ <-> _ => try (lin_split with ; lin_auto)
  | |- _ /\ _ => try (lin_split with ; lin_auto)
  | |- _ [+] _ => try (lin_plus_1 ; lin_auto) ;
                  try (lin_plus_2 ; lin_auto)
  | H : _ [&] _ |- _ => try (lin_left_with_1 H ; lin_auto) ;
                        try (lin_left_with_2 H ; lin_auto)
  | H : _ -> _ |- _ => try (lin_left_imp H with ; lin_auto)
  | H : OC _ |- _ => try (lin_der H ; lin_auto)
  end ;
  fail "Goal not solved"
.
