(*=== Properties of connectives for rewriting ===*)

(* v.0.5.6-1 Olivier Laurent *)

Require Import Coq.Program.Basics.
Require Import Morphisms.
Require Export l2coq.


(*== Linear implication ==*)

Instance lwith_lmap_FORBIDDEN_IN_PROOFS : (* compatibility of & with -o *)
  Proper (impl ++> impl ++> impl) lwith.
Proof.
intros a1 b1 i1 a2 b2 i2 w.
lin_with.
lin_left_with_1 w.
lin_apply1 i1.
clear i2. (* affine *)
lin_assumption.
lin_left_with_2 w.
lin_apply1 i2.
clear i1. (* affine *)
lin_assumption.
Qed.

Instance lplus_lmap_FORBIDDEN_IN_PROOFS : (* compatibility of (+) with -o *)
  Proper (impl ++> impl ++> impl) lplus.
Proof.
intros a1 b1 i1 a2 b2 i2 p.
lin_left_plus p.
lin_plus_1.
lin_apply1 i1.
clear i2. (* affine *)
lin_assumption.
lin_plus_2.
lin_apply1 i2.
clear i1. (* affine *)
lin_assumption.
Qed.

Instance OC_lmap_FORBIDDEN_IN_PROOFS : (* compatibility of ! with -o *)
  Proper (impl ++> impl) OC.
Proof.
intros a o.
unfold impl.
apply oc_fprom_FORBIDDEN_IN_PROOFS.
Qed.

Instance OCit_lmap_FORBIDDEN_IN_PROOFS (n:nat) :
  (* compatibility of !^(n) with setoid_rewrite *)
  Proper (impl ++> impl) (OCit n).
Proof.
induction n ;
intros a o H ;
simpl.
lin_assumption.
lem_apply1 OC_lmap_FORBIDDEN_IN_PROOFS.
lin_apply1 IHn.
lin_assumption.
Qed.

Instance PG_lmap_FORBIDDEN_IN_PROOFS : (* compatibility of § with -o *)
  Proper (impl ++> impl) PG.
Proof.
intros a o.
unfold impl.
apply pg_fprom_FORBIDDEN_IN_PROOFS.
Qed.

Instance PGit_lmap_FORBIDDEN_IN_PROOFS (n:nat) :
  (* compatibility of §^(n) with -o *)
  Proper (impl ++> impl) (PGit n).
Proof.
induction n ;
intros a o H ;
simpl.
lin_assumption.
lem_apply1 PG_lmap_FORBIDDEN_IN_PROOFS.
lin_apply1 IHn.
lin_assumption.
Qed.



(*== Multiplicative equivalence ==*)

Instance lwith_iff_FORBIDDEN_IN_PROOFS : (* compatibility of & with o-o *)
  Proper (iff ==> iff ==> iff) lwith.
Proof.
intros a1 b1 e1 a2 b2 e2.
destruct e1.
destruct e2.
split.
intro w.
lin_with.
lin_left_with_1 w.
lin_apply1 H.
clear H0 H1 H2. (* affine *)
lin_assumption.

lin_left_with_2 w.
lin_apply1 H1.
clear H H0 H2. (* affine *)
lin_assumption.
intro w.
lin_with.
lin_left_with_1 w.
lin_apply1 H0.
clear H H1 H2. (* affine *)
lin_assumption.
lin_left_with_2 w.
lin_apply1 H2.
clear H H0 H1. (* affine *)
lin_assumption.
Qed.

Instance lplus_iff_FORBIDDEN_IN_PROOFS : (* compatibility of (+) with o-o *)
  Proper (iff ==> iff ==> iff) lplus.
Proof.
intros a1 b1 e1 a2 b2 e2.
split.
intro p.
lin_left_plus p.
lin_plus_1.
apply e1. (* affine *)
clear e1 e2. (* affine *)
lin_assumption.
lin_plus_2.
apply e2. (* affine *)
clear e1 e2. (* affine *)
lin_assumption.
intro p.
lin_left_plus p.
lin_plus_1.
apply e1. (* affine *)
clear e1 e2. (* affine *)
lin_assumption.
lin_plus_2.
apply e2. (* affine *)
clear e1 e2. (* affine *)
lin_assumption.
Qed.

Instance OC_iff_FORBIDDEN_IN_PROOFS : (* compatibility of ! with o-o *)
  Proper (iff ==> iff) OC.
Proof.
intros a o H.
lin_left_conj H.
lin_split with H0 ;
apply oc_fprom_FORBIDDEN_IN_PROOFS ;
assumption.
Qed.

Instance OCit_iff_FORBIDDEN_IN_PROOFS (n:nat) :
  (* compatibility of !^(n) with o-&-o *)
  Proper (iff ==> iff) (OCit n).
Proof.
induction n ;
intros a o H ;
simpl.
lin_assumption.
lem_apply1 OC_iff_FORBIDDEN_IN_PROOFS.
lin_apply1 IHn.
lin_assumption.
Qed.

Instance PG_iff_FORBIDDEN_IN_PROOFS : (* compatibility of § with o-o *)
  Proper (iff ==> iff) PG.
Proof.
intros a o H.
lin_left_conj H.
lin_split with H0 ;
apply pg_fprom_FORBIDDEN_IN_PROOFS ;
assumption.
Qed.

Instance PGit_iff_FORBIDDEN_IN_PROOFS (n:nat) :
  (* compatibility of §^(n) with o-o *)
  Proper (iff ==> iff) (PGit n).
Proof.
induction n ;
intros a o H ;
simpl.
lin_assumption.
lem_apply1 PG_iff_FORBIDDEN_IN_PROOFS.
lin_apply1 IHn.
lin_assumption.
Qed.



(*== Additive equivalence ==*)

Instance addiff_impl_subrelation : subrelation addiff impl | 2.
Proof.
intros a b.
unfold addiff.
unfold impl.
intros w Ha.
lin_left_with_1 w.
lin_apply1 w.
lin_assumption.
Qed.

Instance addiff_inverse_impl_subrelation : subrelation addiff (inverse impl) | 2.
Proof.
intros a b.
unfold addiff.
unfold impl.
intros w Ha.
lin_left_with_2 w.
lin_apply1 w.
lin_assumption.
Qed.

Instance addiff_equiv_FORBIDDEN_IN_PROOFS : Equivalence addiff.
Proof.
unfold addiff.
split.

intro a.
lin_with ;
intro Ha ;
lin_assumption.

intros a b e.
lin_with.
lin_left_with_2 e.
lin_assumption.
lin_left_with_1 e.
lin_assumption.

intros a b c e1 e2.
lin_with.
lin_left_with_1 e1.
lin_left_with_1 e2.
intro Ha.
lin_apply1 e2.
lin_apply1 e1.
lin_assumption.
lin_left_with_2 e1.
lin_left_with_2 e2.
intro Hc.
lin_apply1 e1.
lin_apply1 e2.
lin_assumption.
Qed.

Instance lwith_addiff_FORBIDDEN_IN_PROOFS : (* compatibility of & with o-&-o *)
  Proper (addiff ==> addiff ==> addiff) lwith.
Proof.
intros a1 b1 e1 a2 b2 e2.
unfold addiff.
unfold addiff in e1.
unfold addiff in e2.
lin_with.
lin_left_with_1 e1.
lin_left_with_1 e2.
intro w.
lin_with.
lin_left_with_1 w.
lin_apply1 e1.
clear e2. (* affine *)
lin_assumption.
lin_left_with_2 w.
lin_apply1 e2.
clear e1. (* affine *)
lin_assumption.
lin_left_with_2 e1.
lin_left_with_2 e2.
intro w.
lin_with.
lin_left_with_1 w.
lin_apply1 e1.
clear e2. (* affine *)
lin_assumption.
lin_left_with_2 w.
lin_apply1 e2.
clear e1. (* affine *)
lin_assumption.
Qed.

Instance lplus_addiff_FORBIDDEN_IN_PROOFS : (* compatibility of (+) with o-&-o *)
  Proper (addiff ==> addiff ==> addiff) lplus.
Proof.
intros a1 b1 e1 a2 b2 e2.
split.
intro p.
lin_left_plus p.
lin_plus_1.
apply e1. (* affine *)
clear e1 e2. (* affine *)
lin_assumption.
lin_plus_2.
apply e2. (* affine *)
clear e1 e2. (* affine *)
lin_assumption.
intro p.
lin_left_plus p.
lin_plus_1.
apply e1. (* affine *)
clear e1 e2. (* affine *)
lin_assumption.
lin_plus_2.
apply e2. (* affine *)
clear e1 e2. (* affine *)
lin_assumption.
Qed.

Instance OC_addiff_FORBIDDEN_IN_PROOFS : (* compatibility of ! with o-&-o *)
  Proper (addiff ==> addiff) OC.
Proof.
intros a o H.
lin_left_conj H.
lin_split with H0 ;
apply oc_fprom_FORBIDDEN_IN_PROOFS ;
assumption.
Qed.

Instance OCit_addiff_FORBIDDEN_IN_PROOFS (n:nat) :
  (* compatibility of !^(n) with o-&-o *)
  Proper (addiff ==> addiff) (OCit n).
Proof.
induction n ;
intros a o H ;
simpl.
lin_assumption.
lem_apply1 OC_addiff_FORBIDDEN_IN_PROOFS.
lin_apply1 IHn.
lin_assumption.
Qed.

Instance PG_addiff_FORBIDDEN_IN_PROOFS : (* compatibility of § with o-&-o *)
  Proper (addiff ==> addiff) PG.
Proof.
intros a o H.
lin_left_conj H.
lin_split with H0 ;
apply pg_fprom_FORBIDDEN_IN_PROOFS ;
assumption.
Qed.

Instance PGit_addiff_FORBIDDEN_IN_PROOFS (n:nat) :
  (* compatibility of §^(n) with o-&-o *)
  Proper (addiff ==> addiff) (PGit n).
Proof.
induction n ;
intros a o H ;
simpl.
lin_assumption.
lem_apply1 PG_addiff_FORBIDDEN_IN_PROOFS.
lin_apply1 IHn.
lin_assumption.
Qed.



