(*== Classical mutliplicative linear logic in Coq ==*)

(* v.0.1 Olivier Laurent *)

Require Import l2coq.


(*= Preliminaries =*)

Parameter Concl : Prop.

Definition CLL : Prop -> Prop := fun A => (A -> Concl).


(*= Identity =*)

Definition clorth (A : Prop) : Prop := (* Linear negation *)
  A -> Concl.
Notation "{~} A" := (clorth A) (at level 35).

Ltac clin_ax := (* Axiom rule *)
  match goal with
  | H1 : {~} ?A, H2 : ?A |- _ => lin_hyp_apply1 H1 in H2 ;
                                 lin_assumption
  end
  || fail "Invalid axiom rule".

Tactic Notation (* Cut rule *)
  "clin_cut" constr(c) "as" simple_intropattern(h) "with" hyp_list(G) :=
    lin_bin_split_general ltac:(cut ({~}{~} c))
       ltac:(idtac) ltac:(clear - G) ltac:(intro h ; lin_apply1 h ; intro h)
       ltac:(idtac) ltac:(clear G) ltac:(intro h).


(*= Multiplicatives =*)

Inductive clbot : Prop := (* Bottom *)
| clbot_c : clbot.
Notation "{b}" := clbot.

Definition clone := Concl. (* One *)
Notation "{1}" := clone.

Inductive clpar (A B : Prop) : Prop := (* Par *)
| clpar_c : A -> B -> clpar A B.
Notation "A  {P}  B" := (clpar A B) (at level 46).

Definition cltens A B := (* Tensor *)
  (A -> Concl) -> (B -> Concl) -> Concl.
Notation "A  {x}  B" := (cltens A B) (at level 45).

Ltac clin_bot H := (* Bottom rule *)
  lazymatch type of H with
  | {b} => clear H
  end
  || fail "Hypothesis is not bottom".

Ltac clin_one := (* One rule *)
  lin_assumption
  || fail "Invalid one rule".

Ltac clin_par H := (* Par rule *)
  lazymatch type of H with
  | ?A {P} ?B => let H1 := fresh H in
                 let H2 := fresh H in
                 destruct H as [H1 H2]
  end
  || fail "Hypothesis is not a par".

Tactic Notation "clin_tens" hyp(H) "with" hyp_list(G) := (* Tensor rule *)
  lazymatch type of H with
  | ?A {x} ?B => lin_bin_split
                   ltac:(simple apply H)
                   ltac:(clear - G) ltac:(clear G ; clear H) ;
                 intro H
  end
  || fail "Hypothesis is not a tensor".


