(*= Church encodings in Light Linear Logic =*)

(* v.0.1 Olivier Laurent *)

Require Import stdlib_lll2coq. (* v.0.5.6-1 *)

Definition CLNat := forall X : Prop, !(X -> X) -> §(X -> X).

Fact clzero : CLNat.
Proof.
intros X f.
lin_weak f.
lin_mfprom_pg.
intro x.
lin_assumption.
(* llcheck
LLL.
llcheck *)
Qed.

Fact clone : CLNat.
Proof.
intros X f.
lin_mfprom_pg.
lin_assumption.
(* llcheck
LLL.
llcheck *)
Qed.

Fact cltwo : CLNat.
Proof.
intros X f.
lin_contr f.
lin_mfprom_pg.
intro x.
lin_apply1 f0.
lin_apply1 f.
lin_assumption.
(* llcheck
LLL.
llcheck *)
Qed.

Fact cladd : CLNat -> CLNat -> CLNat.
Proof.
intros m n.
intros X f.
lin_contr f.
lin_hyp_apply1 m in f.
lin_hyp_apply1 n in f0.
lin_mfprom_pg.
intro x.
lin_apply1 f0.
lin_apply1 f.
lin_assumption.
(* llcheck
LLL.
llcheck *)
Qed.

Fact clsucc : CLNat -> CLNat.
Proof.
intro n.
lem_apply2 cladd with.
lem_apply0 clone.
lin_assumption.
(* llcheck
LLL.
llcheck *)
Qed.

Ltac liter n :=
  lazymatch goal with
  | |- OC ?A -> PG ?A => let H := fresh in
                         specialize (n A) ;
                         intro H ;
                         lem_hyp_apply1 pg_der in H ;
                         revert H ;
                         lem_apply1 mon_pg_lmap ;
                         lin_apply1 n ;
                         lin_ufprom_oc
  | _                 => fail "Conclusion is not of the shape !A -> §A"
  end.

Fact clmult : CLNat -> !CLNat -> §CLNat.
Proof.
intros m n.
specialize (m CLNat).
lin_cut (§CLNat) as Z with m n.
revert Z.
lem_apply1 mon_pg_lmap.
lin_apply1 m.
lin_fprom_oc.
lem_apply1 cladd.
lin_assumption.
lin_mfprom_pg.
lem_apply0 clzero.
(* llcheck
LLL.
llcheck *)
Qed.

Fact clnat_coer : CLNat -> §CLNat.
Proof.
intro n.
lin_cut (!CLNat) as i with n.
revert i.
liter n.
lem_apply0 clsucc.
lin_ufprom_oc.
lem_apply0 clzero.
(* llcheck
LLL.
llcheck *)
Qed.


Definition CLBin := forall X : Prop, !(X -> X) -> !(X -> X) -> §(X -> X).

Fact clnil : CLBin.
Proof.
intros X c0 c1.
lin_weak c0.
lin_weak c1.
lin_mfprom_pg.
intro n.
lin_assumption.
(* llcheck
LLL.
llcheck *)
Qed.

Fact clcons0 : CLBin -> CLBin.
Proof.
intros l X c0 c1.
lin_contr c0.
specialize (l X).
lin_left_imp l with c1 c2.
lin_left_imp l with c2.
lin_mfprom_pg.
intro n.
lin_apply1 l.
lin_apply1 c2.
lin_assumption.
lin_assumption.
lin_assumption.
(* llcheck
LLL.
llcheck *)
Qed.

Fact clcons1 : CLBin -> CLBin.
Proof.
intros l X c0 c1.
lin_contr c1.
specialize (l X).
lin_left_imp l with c1 c2.
lin_left_imp l with c2.
lin_mfprom_pg.
intro n.
lin_apply1 l.
lin_apply1 c2.
lin_assumption.
lin_assumption.
lin_assumption.
(* llcheck
LLL.
llcheck *)
Qed.

Fact clappend : CLBin -> CLBin -> CLBin.
Proof.
intros l1 l2.
intros X c0 c1.
lin_contr c0.
lin_contr c1.
specialize (l1 X).
specialize (l2 X).
lin_left_imp l1 with l2 c1 c2 c3.
lin_left_imp l1 with l2 c2 c3.
lin_left_imp l2 with l1 c2.
lin_left_imp l2 with l1.
lin_mfprom_pg.
intro x.
lin_apply1 l2.
lin_apply1 l1.
lin_assumption.
lin_assumption.
lin_assumption.
lin_assumption.
lin_assumption.
(* llcheck
LLL.
llcheck *)
Qed.

Fact cllength : CLBin -> CLNat.
Proof.
intro l.
intros X s.
lin_contr s.
lin_apply2 l with s ;
lin_assumption.
(* llcheck
LLL.
llcheck *)
Qed.

Ltac liter2 l :=
  lazymatch goal with
  | |- OC ?A -> PG ?A => let H := fresh in
                         specialize (l A) ;
                         intro H ;
                         lem_hyp_apply1 pg_der in H ;
                         revert H ;
                         lem_apply1 mon_pg_lmap ;
                         (lin_apply2 l with ; lin_ufprom_oc)
  | _                 => fail "Conclusion is not of the shape !A -> §A"
  end.

Fact clbin_coer : CLBin -> §CLBin.
Proof.
intros l.
lin_cut (!CLBin) as i with l.
revert i.
liter2 l.
lem_apply0 clcons0.
lem_apply0 clcons1.
lin_ufprom_oc.
lem_apply0 clnil.
(* llcheck
LLL.
llcheck *)
Qed.


