(*=== Standard library for l2coq ===*)
(*== ILL ==*)

(* v.0.5.6-1 Olivier Laurent *)

Require Export l2coq.
Require Export stdlib_mall2coq.
Require Import l2coq_auto. (* v.0.4 *)
Require Import Arith.

Local Open Scope mll_scope.
Local Open Scope all_scope.


(*== Exponentials ==*)

Lemma ax_exp_oc : forall a : Prop, !a -> !a.
Proof.
intros a o.
lin_prom.
lin_assumption_der.
(* llcheck
LL.
llcheck *)
Qed.


(*== Exponentials and quantifiers ==*)

Lemma oc_forall : forall T0, forall a : T0 -> Prop,
  !(forall x : T0, a x) -> forall x : T0, !(a x).
Proof.
intros T0 a Ho x.
lin_prom.
lin_der Ho.
lin_apply0 Ho.
(* llcheck
LL.
llcheck *)
Qed.

Lemma exists_oc : forall T0, forall a : T0 -> Prop,
  (exists x : T0, !(a x)) -> !(exists x : T0, a x).
Proof.
intros T0 a e.
destruct e as [x].
lin_prom.
exists x.
lin_assumption_der.
(* llcheck
LL.
llcheck *)
Qed.


(*== Exponentials and multiplicatives ==*)

Lemma mon_oc_one : True -> !True.
Proof.
lin_rever.
(* llcheck
LL.
llcheck *)
Qed.

Lemma mon_oc_lmap : forall a b : Prop, !(a -> b) -> !a -> !b.
Proof.
lin_auto.
(* llcheck
LL.
llcheck *)
Qed.

Lemma mon_oc_tens : forall a b : Prop, (!a /\ !b) -> !(a /\ b).
Proof.
lin_rever.
lin_split with H0 ;
lin_assumption_der.
(* llcheck
LL.
llcheck *)
Qed.


(*== Exponentials and additives ==*)

Lemma oc_with :
  forall a b : Prop, !(a [&] b) -> !a [&] !b.
Proof.
lin_auto.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_oc :
  forall a b : Prop, !a [+] !b -> !(a [+] b).
Proof.
lin_auto.
(* llcheck
LL.
llcheck *)
Qed.


(*== Exponentials, multiplicatives and additives ==*)

Lemma iso_mae_unit : ![T] o-&-o True.
Proof.
lin_rever.
(* llcheck
LL.
llcheck *)
Qed.

Lemma iso_mae :
  forall a b : Prop, !(a [&] b) o-&-o (!a /\ !b).
Proof.
(*
intros a b.
lin_split_equiv.

intro w.
lin_contr w.
lin_split with w.
lin_prom.
lin_der w.
lin_left_with_1 w.
lin_assumption.
lin_prom.
lin_der w0.
lin_left_with_2 w0.
lin_assumption.

intro t.
lin_left_conj t.
lin_prom.
lin_with.
lin_weak t1.
lin_der t0.
lin_assumption.
lin_weak t0.
lin_der t1.
lin_assumption.
*)
lin_rever.
lin_contr H.
lin_split with H ;
lin_auto.
(* llcheck
LL.
llcheck *)
Qed.


(*== Second order encodings ==*)

Lemma sndenc_plus :
  forall a b : Prop, (a [+] b) o-&-o forall X : Prop, !(a -> X) -> !(b -> X) -> X.
Proof.
(*
intros a b.
lin_split_equiv.
intros p x ia ib.
lin_left_plus p.
lin_der ia.
lin_weak ib.
lin_apply1 ia.
lin_assumption.
lin_der ib.
lin_weak ia.
lin_apply1 ib.
lin_assumption.

intros s.
specialize s with (a [+] b).
lin_apply2 s with.
lin_prom.
intro ha.
lin_plus_1.
lin_assumption.
lin_prom.
intro hb.
lin_plus_2.
lin_assumption.
*)
lin_rever.
lin_der H0.
lin_apply1 H0.
lin_assumption_oc.
lin_der H1.
lin_apply1 H1.
lin_assumption_oc.
lin_apply2 H with ;
lin_auto.
(* llcheck
LL.
llcheck *)
Qed.


(*== Iterated exponentials ==*)

Lemma ocit_der : forall n : nat, forall a : Prop, !^(n) a -> a.
Proof.
intros n a.
induction n ;
simpl.
intro ; lin_assumption.
intro i.
lin_apply1 IHn.
lin_der i.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Ltac lin_der_ocit H := lem_hyp_apply1 ocit_der in H.
Ltac lin_right_der_ocit := lem_apply1 ocit_der.

Lemma ocit_dig : forall n : nat, forall a : Prop, !a -> !^(n) a.
Proof.
intros n a o.
lin_prom_it.
lin_der o.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Ltac lin_dig_ocit H := lem_hyp_apply1 ocit_dig in H.
Ltac lin_right_dig_ocit := lem_apply1 ocit_dig.

Lemma ocit_it_it : forall m n : nat, forall a : Prop, !^(S m) a -> !^(n) a.
Proof.
simpl.
intros m n a om.
lem_apply1 ocit_dig.
revert om.
rewrite ococit_ocitoc.
lem_apply0 ocit_der.
(* llcheck
LL.
llcheck *)
Qed.

Lemma ocit_S_S : forall m n : nat, forall a : Prop, !^(S m) a o-&-o !^(S n) a.
intros m n a.
lin_split_equiv ;
lem_apply0 ocit_it_it.
(* llcheck
LL.
llcheck *)
Qed.


Local Close Scope mll_scope.
Local Close Scope all_scope.
