(*=== Standard library for l2coq ===*)
(*== ILLL ==*)

(* v.0.5.6-1 Olivier Laurent *)

Require Export l2coq.
Require Export stdlib_mall2coq.
Require Import l2coq_auto.

Local Open Scope mll_scope.
Local Open Scope all_scope.

(*== Exponentials ==*)

Lemma ax_exp_ocf : forall a : Prop, !a -> !a.
Proof.
intros a O.
lin_fprom_oc.
lin_assumption.
(* llcheck
LLL.
ELL.
SLL.
LLax.
llcheck *)
Qed.


(*== Exponentials and quantifiers ==*)

Lemma oc_forall : forall T0, forall a : T0 -> Prop,
  !(forall x : T0, a x) -> forall x : T0, !(a x).
Proof.
intros T0 a Ho x.
lin_fprom_oc.
lin_apply0 Ho.
(* llcheck
LLL.
ELL.
SLL.
LLax.
llcheck *)
Qed.

Lemma exists_oc : forall T0, forall a : T0 -> Prop,
  (exists x : T0, !(a x)) -> !(exists x : T0, a x).
Proof.
intros T0 a He.
destruct He as [x].
lin_fprom_oc.
exists x.
lin_assumption.
(* llcheck
LLL.
ELL.
SLL.
LLax.
llcheck *)
Qed.


(*== Exponentials and multiplicatives ==*)

Lemma mon_oc_one : True -> !True.
Proof.
(*
intro O.
lin_left_one O.
lin_ufprom_oc.
lin_one.
*)
lin_rever_lll.
(* llcheck
LLL.
ELL.
SLL.
LLax.
llcheck *)
Qed.


(*== Exponentials and additives ==*)

Lemma oc_with :
  forall a b : Prop, !(a [&] b) -> !a [&] !b.
Proof.
(*
intros a b.
intro O.
lin_with ;
lin_fprom_oc.
lin_left_with_1 O.
lin_assumption.
lin_left_with_2 O.
lin_assumption.
*)
lin_auto_lll.
(* llcheck
LLL.
ELL.
SLL.
LLax.
llcheck *)
Qed.

Lemma plus_oc :
  forall a b : Prop, !a [+] !b -> !(a [+] b).
Proof.
(*
intros a b.
intro P.
lin_left_plus P ;
lin_fprom_oc.
lin_plus_1.
lin_assumption.
lin_plus_2.
lin_assumption.
*)
lin_auto_lll.
(* llcheck
LLL.
ELL.
SLL.
LLax.
llcheck *)
Qed.


(*== Exponentials, multiplicatives and additives ==*)

Lemma imp_mae :
  forall a b : Prop, !(a [&] b) -> !a /\ !b.
Proof.
intros a b W.
lin_contr W.
(*
lin_split with W.
lin_fprom_oc.
lin_left_with_1 W.
lin_assumption.
lin_fprom_oc.
lin_left_with_2 W0.
lin_assumption.
*)
lin_split with W ;
lin_auto_lll.
(* llcheck
LLL.
ELL.
Fail SLL.
LLax.
llcheck *)
Qed.

Lemma iso_mae_unit : ![T] o-&-o True.
Proof.
(*
lin_split_equiv.
intro t.
lin_weak t.
lin_one.
intro O.
lin_left_one O.
lin_ufprom_oc.
lin_top.
*)
lin_rever_lll.
(* llcheck
LLL.
ELL.
LLax.
llcheck *)
Qed.


(*== Paragraph ==*)

Lemma ax_exp_pgf : forall a : Prop, §a -> §a.
Proof.
intros a O.
lin_mfprom_pg.
lin_assumption.
(* llcheck
LLL.
Fail ELL.
Fail SLL.
Fail LLax.
llcheck *)
Qed.


(*== Paragraph and quantifiers ==*)

Lemma pg_forall : forall T0, forall a : T0 -> Prop,
  §(forall x : T0, a x) -> forall x : T0, §(a x).
Proof.
(*
intros T0 a Ho x.
lin_mfprom_pg.
specialize Ho with x.
lin_assumption.
*)
lin_rever_lll.
lin_apply0 H.
(* llcheck
LLL.
Fail ELL.
Fail SLL.
Fail LLax.
llcheck *)
Qed.

Lemma exists_pg : forall T0, forall a : T0 -> Prop,
  (exists x : T0, §(a x)) -> §(exists x : T0, a x).
Proof with lin_rever_lll.
(*
intros T0 a He.
destruct He as [x].
lin_mfprom_pg.
exists x.
lin_assumption.
*)
lin_rever_lll.
exists x...
(* llcheck
LLL.
Fail ELL.
Fail SLL.
Fail LLax.
llcheck *)
Qed.


(*== Paragraph and multiplicatives ==*)

Lemma mon_pg_lmap : forall a b : Prop, §(a -> b) -> §a -> §b.
Proof.
(*
intros a b Hi Ha.
lin_mfprom_pg.
lin_apply1 Hi.
lin_assumption.
*)
lin_auto_lll.
(* llcheck
LLL.
Fail ELL.
Fail SLL.
Fail LLax.
llcheck *)
Qed.

Lemma mon_pg_tens : forall a b : Prop, (§a /\ §b) -> §(a /\ b).
Proof with lin_rever_lll.
(*
intros a b t.
lin_left_conj t.
lin_mfprom_pg.
lin_split with t0 ;
lin_assumption.
*)
lin_rever_lll.
lin_split with H0...
(* llcheck
LLL.
Fail ELL.
Fail SLL.
Fail LLax.
llcheck *)
Qed.

Lemma mon_pg_one : True -> §True.
Proof.
(*
intro O.
lin_left_one O.
lin_mfprom_pg.
lin_one.
*)
lin_rever_lll.
(* llcheck
LLL.
Fail ELL.
Fail SLL.
Fail LLax.
llcheck *)
Qed.


(*== Paragraph and additives ==*)

Lemma pg_with :
  forall a b : Prop, §(a [&] b) -> §a [&] §b.
Proof.
(*
intros a b.
intro O.
lin_with ;
lin_mfprom_pg.
lin_left_with_1 O.
lin_assumption.
lin_left_with_2 O.
lin_assumption.
*)
lin_auto_lll.
(* llcheck
LLL.
Fail ELL.
Fail SLL.
Fail LLax.
llcheck *)
Qed.

Lemma plus_pg :
  forall a b : Prop, §a [+] §b -> §(a [+] b).
Proof.
(*
intros a b.
intro P.
lin_left_plus P ;
lin_mfprom_pg.
lin_plus_1.
lin_assumption.
lin_plus_2.
lin_assumption.
*)
lin_auto_lll.
(* llcheck
LLL.
Fail ELL.
Fail SLL.
Fail LLax.
llcheck *)
Qed.


(*== Paragraph and exponentials ==*)

Lemma oc_pg : forall a : Prop, !a -> §a.
Proof.
(*
intros a O.
lin_mfprom_pg.
lin_assumption.
*)
lin_auto_lll.
(* llcheck
LLL.
Fail ELL.
Fail SLL.
Fail LLax.
llcheck *)
Qed.


Local Close Scope mll_scope.
Local Close Scope all_scope.


