(*=== Standard library for l2coq ===*)
(*== IMALL ==*)

(* v.0.5.6-1 Olivier Laurent *)

Require Export l2coq.
Require Import l2coq_auto. (* v.0.4 *)

Local Open Scope mll_scope.
Local Open Scope all_scope.

(*== Quantifiers ==*)

Lemma forall_forall_r : forall T1 T2, forall a : T1 -> T2 -> Prop,
  (forall x1 : T1, forall x2 : T2, a x1 x2)
    -> forall x2 : T2, forall x1 : T1, a x1 x2.
Proof.
intros T1 T2 a H x2 x1.
(*
specialize H with x1 x2.
lin_assumption.
*)
lin_apply0 H.
(* llcheck
LL.
llcheck *)
Qed.

Lemma forall_forall : forall T1 T2, forall a : T1 -> T2 -> Prop,
  (forall x1 : T1, forall x2 : T2, a x1 x2)
    o-&-o forall x2 : T2, forall x1 : T1, a x1 x2.
Proof.
intros T1 T2 a.
lin_split_equiv ;
lem_apply0 forall_forall_r.
(* llcheck
LL.
llcheck *)
Qed.

Lemma empty_forall : forall T0, forall a : Prop, a -> forall x : T0, a.
Proof.
(*
intros T0 a Ha x.
lin_assumption.
*)
lin_rever_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma neforall_empty :
  forall T0, (exists c : T0, True) -> forall a : Prop, (forall x : T0, a) -> a.
Proof.
(*
intros T0 E.
destruct E as [c O].
lin_left_one O.
intros a F.
lin_hyp_apply1 F in c.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma forall_empty_prop :
  forall a : Prop, (forall x : Prop, a) o-&-o a.
Proof.
intro a.
lin_split_equiv.
lem_apply1 neforall_empty.
exists True.
lin_one.
lem_apply0 empty_forall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma forallforall_forall :
  forall T0, forall a : T0 -> T0 -> Prop,
    (forall x : T0, forall y : T0, a x y) -> forall x : T0, a x x.
Proof.
intros T0 a FF x.
(*
specialize FF with x x.
lin_assumption.
*)
lin_apply0 FF.
(* llcheck
LL.
llcheck *)
Qed.

Lemma exists_exists_r : forall T1 T2, forall a : T1 -> T2 -> Prop,
  (exists x1 : T1, exists x2 : T2, a x1 x2)
    -> exists x2 : T2, exists x1 : T1, a x1 x2.
Proof.
intros T1 T2 a H.
destruct H as [x1].
destruct H as [x2].
exists x2.
exists x1.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma exists_exists : forall T1 T2, forall a : T1 -> T2 -> Prop,
  (exists x1 : T1, exists x2 : T2, a x1 x2)
    o-&-o exists x2 : T2, exists x1 : T1, a x1 x2.
Proof.
intros T1 T2 a.
lin_split_equiv ;
lem_apply0 exists_exists_r.
(* llcheck
LL.
llcheck *)
Qed.

Lemma exists_empty : forall T0, forall a : Prop, (exists x : T0, a) -> a.
Proof.
(*
intros T0 a H.
destruct H as [x].
lin_assumption.
*)
lin_rever_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma empty_neexists :
  forall T0, (exists c : T0, True) -> forall a : Prop, a -> exists x : T0, a.
Proof.
(*
intros T0 E.
destruct E as [c O].
lin_left_one O.
intros a Ha.
exists c.
lin_assumption.
*)
lin_rever_mall.
exists x.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma exists_empty_prop :
  forall a : Prop, (exists x : Prop, a) o-&-o a.
Proof.
intro a.
lin_split_equiv.
lem_apply0 exists_empty.
lem_apply1 empty_neexists.
exists True.
lin_one.
(* llcheck
LL.
llcheck *)
Qed.

Lemma exists_existsexists :
  forall T0, forall a : T0 -> T0 -> Prop,
    (exists x : T0, a x x) -> exists x : T0, exists y : T0, a x y.
Proof.
intros T0 a E.
destruct E as [x Ha].
exists x.
exists x.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.


(*== Multiplicatives ==*)

Lemma ax_exp_one : True -> True.
Proof.
intro O.
lin_left_one O.
lin_one.
(* llcheck
LL.
llcheck *)
Qed.

Lemma ax_exp_tens : forall a b : Prop, a /\ b -> a /\ b.
Proof.
intros a b t.
lin_left_conj t.
lin_split with t0 ;
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma ax_exp_lmap : forall a b : Prop, (a -> b) -> a -> b.
Proof.
intros a b I.
intro Ha.
lin_left_imp I with ;
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma tens_comm : forall a b : Prop, (a /\ b) o-&-o (b /\ a).
Proof.
intros a b.
lin_split_equiv ;
intro t ;
lin_left_conj t ;
lin_split with t1 ;
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma tens_assoc : forall a b c : Prop, ((a /\ b) /\ c) o-&-o (a /\ (b /\ c)).
Proof.
intros a b c.
lin_split_equiv ;
intro t ;
lin_left_conj t.

lin_left_conj t0.
lin_split with t2.
lin_assumption.
lin_split with t3 ;
lin_assumption.

lin_left_conj t1.
lin_split with t0 t2.
lin_split with t0 ;
lin_assumption.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma tens_unit_r : forall a : Prop, (a /\ True) o-&-o a.
Proof with lin_rever_mall.
(*
intro a.
lin_split_equiv.

intro t.
lin_left_conj t.
lin_left_one t1.
lin_assumption.

intro Ha.
lin_split with Ha.
lin_assumption.
lin_one.
*)
lin_rever_mall.
lin_split with H...
(* llcheck
LL.
llcheck *)
Qed.

Lemma tens_unit_l : forall a : Prop, (True /\ a) o-&-o a.
Proof.
(*
intro a.
lin_split_equiv.

intro t.
lin_left_conj t.
lin_left_one t0.
lin_assumption.

intro Ha.
lin_split with.
lin_one.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma lmap_comm : forall a b c : Prop, (a -> b -> c) o-&-o (b -> a -> c).
Proof.
intros a b c.
lin_split_equiv ;
intros I H1 H2 ;
lin_apply2 I with H2 ;
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma lmap_unit_l : forall a : Prop, (True -> a) o-&-o a.
Proof.
(*
intro a.
lin_split_equiv.

intro I.
lin_apply1 I.
lin_one.

intros Ha O.
lin_left_one O.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma lin_curry : forall a b c : Prop, ((a /\ b) -> c) o-&-o (a -> b -> c).
Proof.
intros a b c.
lin_split_equiv.
intros t Ha Hb.
lin_apply1 t.
lin_split with Ha ;
lin_assumption.

intros I t.
lin_left_conj t.
lin_apply2 I with t0 ;
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma lin_distr_r : forall a b c : Prop, ((a -> b) /\ c) -> a -> b /\ c.
Proof with lin_rever_mall.
(*
intros a b c H Ha.
lin_left_conj H.
lin_split with H0 Ha.
lin_apply1 H0.
lin_assumption.
lin_assumption.
*)
lin_rever_mall.
lin_left_imp H1 with H2...
lin_split with H1...
(* llcheck
LL.
llcheck *)
Qed.

Lemma lin_distr_l : forall a b c : Prop, (a /\ (b -> c)) -> b -> a /\ c.
Proof with lin_rever_mall.
(*
intros a b c H Hb.
lin_left_conj H.
lin_split with H0.
lin_assumption.
lin_apply1 H1.
lin_assumption.
*)
lin_rever_mall.
lin_left_imp H2 with H1...
lin_split with H1...
(* llcheck
LL.
llcheck *)
Qed.

Lemma lin_tens_lmaps :
  forall a b c d : Prop, ((a -> c) /\ (b -> d)) -> (a /\ b) -> (c /\ d).
Proof.
intros a b c d t.
lem_apply1 lin_curry.
intro Ha.
lem_apply1 lin_distr_l.
revert Ha.
lem_apply1 lin_distr_r.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.


(*== Multiplicatives and quantifiers ==*)

Lemma tens_exists :
  forall T0, forall a : Prop, forall b : T0 -> Prop,
    (a /\ exists x : T0, b x) o-&-o exists x : T0, (a /\ b x).
Proof with lin_rever_mall.
(*
intros T0 a b.
lin_split_equiv ;
intro H.
lin_left_conj H.
destruct H1 as [x].
exists x.
lin_split with H0 ;
lin_assumption.

destruct H as [x].
lin_left_conj H.
lin_split with H0.
lin_assumption.
exists x.
lin_assumption.
*)
lin_rever_mall.
exists x.
lin_split with H0...
lin_split with H0...
exists x...
(* llcheck
LL.
llcheck *)
Qed.

Lemma lmap_forall :
  forall T0, forall a : Prop, forall b : T0 -> Prop,
    (a -> forall x : T0, b x) o-&-o forall x : T0, (a -> b x).
Proof.
(*
intros T0 a b.
lin_split_equiv ;
intro H.
intros x Ha.
lin_apply1 H.
lin_assumption.
intros Ha X.
lin_apply1 H.
lin_assumption.
*)
lin_rever_mall ;
lin_apply1 H ;
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma tens_forall :
  forall T0, forall a : Prop, forall b : T0 -> Prop,
    (a /\ forall x : T0, b x) -> forall x : T0, (a /\ b x).
Proof with lin_rever_mall.
(*
intros T0 a b HT x.
lin_left_conj HT.
lin_split with HT0.
lin_assumption.
lin_apply0 HT1.
*)
lin_rever_mall.
lin_split with H0...
lin_apply0 H1.
(* llcheck
LL.
llcheck *)
Qed.

Lemma exists_lmap :
  forall T0, forall a : Prop, forall b : T0 -> Prop,
    (exists x : T0, (a -> b x)) -> a -> exists x : T0 , b x.
Proof.
intros T0 a b E Ha.
destruct E as [x I].
exists x.
lin_apply1 I.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma exists_lmap_forall_left :
  forall T0, forall a : T0 -> Prop, forall b : Prop,
    (exists x : T0, (a x -> b)) -> (forall x : T0, a x) -> b.
Proof.
(*
intros T0 a b E F.
destruct E as [x I].
lin_apply1 I.
specialize F with x.
lin_assumption.
*)
lin_rever_mall.
lin_apply1 H.
lin_apply0 H0.
(* llcheck
LL.
llcheck *)
Qed.


(*== Additives ==*)

Lemma ax_exp_top : [T] -> [T].
Proof.
intro t.
lin_top.
(* llcheck
LL.
llcheck *)
Qed.

Lemma ax_exp_with : forall a b : Prop, a [&] b -> a [&] b.
Proof.
intros a b W.
lin_with.
lin_left_with_1 W.
lin_assumption.
lin_left_with_2 W.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma ax_exp_zero : [0] -> [0].
Proof.
intro Z.
lin_left_zero Z.
(* llcheck
LL.
llcheck *)
Qed.

Lemma ax_exp_plus : forall a b : Prop, a [+] b -> a [+] b.
Proof with lin_rever_mall.
(*
intros a b P.
lin_left_plus P.
lin_plus_1.
lin_assumption.
lin_plus_2.
lin_assumption.
*)
intros a b P...
lin_plus_1...
lin_plus_2...
(* llcheck
LL.
llcheck *)
Qed.

Lemma with_comm : forall a b : Prop, (a [&] b) o-&-o (b [&] a).
Proof.
(*
intros a b.
lin_with ;
intro W ;
(lin_with ;
[
lin_left_with_2 W ;
lin_assumption
|
lin_left_with_1 W ;
lin_assumption
]).
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma with_assoc :
  forall a b c : Prop, ((a [&] b) [&] c) o-&-o (a [&] (b [&] c)).
Proof.
(*
intros a b c.
lin_split_equiv ;
intro W ;
lin_with.

lin_left_with_1 W.
lin_left_with_1 W.
lin_assumption.
lin_with.
lin_left_with_1 W.
lin_left_with_2 W.
lin_assumption.
lin_left_with_2 W.
lin_assumption.

lin_with.
lin_left_with_1 W.
lin_assumption.
lin_left_with_2 W.
lin_left_with_1 W.
lin_assumption.
lin_left_with_2 W.
lin_left_with_2 W.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma with_unit_r : forall a : Prop, (a [&] [T]) o-&-o a.
Proof.
(*
intro a.
lin_split_equiv.
intro W.
lin_left_with_1 W.
lin_assumption.
intro Ha.
lin_with.
lin_assumption.
lin_top.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma with_unit_l : forall a : Prop, ([T] [&] a) o-&-o a.
Proof.
(*
intro a.
lin_split_equiv.
intro W.
lin_left_with_2 W.
lin_assumption.
intro Ha.
lin_with.
lin_top.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_comm : forall a b : Prop, (a [+] b) o-&-o (b [+] a).
Proof.
(*
intros a b.
lin_split_equiv ;
intro P ;
(lin_left_plus P ;
[
lin_plus_2 ;
lin_assumption
|
lin_plus_1 ;
lin_assumption
]).
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_assoc :
  forall a b c : Prop, ((a [+] b) [+] c) o-&-o (a [+] (b [+] c)).
Proof.
(*
intros a b c.
lin_split_equiv ;
intro P ;
lin_left_plus P.

lin_left_plus P.
lin_plus_1.
lin_assumption.
lin_plus_2.
lin_plus_1.
lin_assumption.
lin_plus_2.
lin_plus_2.
lin_assumption.

lin_plus_1.
lin_plus_1.
lin_assumption.
lin_left_plus P.
lin_plus_1.
lin_plus_2.
lin_assumption.
lin_plus_2.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_unit_r : forall a : Prop, (a [+] [0]) o-&-o a.
Proof.
(*
intro a.
lin_split_equiv.

intro P.
lin_left_plus P.
lin_assumption.
lin_left_zero P.

intro Ha.
lin_plus_1.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_unit_l : forall a : Prop, ([0] [+] a) o-&-o a.
Proof.
(*
intro a.
lin_split_equiv.

intro P.
lin_left_plus P.
lin_left_zero P.
lin_assumption.

intro Ha.
lin_plus_2.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma top_terminal : forall a : Prop, a -> [T].
Proof.
(*
intros a Ha.
lin_top.
*)
lin_rever_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma zero_initial : forall a : Prop, [0] -> a.
Proof.
(*
intros a Z.
lin_left_zero Z.
*)
lin_rever_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma with_diag : forall a : Prop, a -> a [&] a.
Proof.
(*
intros a Ha.
lin_with ;
lin_assumption.
*)
lin_rever_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma with_proj_1 : forall a b : Prop, a [&] b -> a.
(*
intros a b W.
lin_left_with_1 W.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma with_proj_2 : forall a b : Prop, a [&] b -> b.
(*
intros a b W.
lin_left_with_2 W.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_codiag : forall a : Prop, a [+] a -> a.
(*
intros a P.
lin_left_plus P ;
lin_assumption.
*)
lin_rever_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_inj_1 : forall a b : Prop, a -> a [+] b.
(*
intros a b Ha.
lin_plus_1.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_inj_2 : forall a b : Prop, b -> a [+] b.
(*
intros a b Ha.
lin_plus_2.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_with :
  forall a b c : Prop, a [+] (b [&] c) -> (a [+] b) [&] (a [+] c).
Proof.
(*
intros a b c P.
lin_with.
lin_left_plus P.
lin_plus_1.
lin_assumption.
lin_left_with_1 P.
lin_plus_2.
lin_assumption.

lin_left_plus P.
lin_plus_1.
lin_assumption.
lin_left_with_2 P.
lin_plus_2.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.


(*== Additives and quantifiers ==*)

Lemma top_forall :
  forall T0, [T] o-&-o forall x : T0, [T].
Proof.
(*
intros T0.
lin_split_equiv.
intros H x.
lin_top.
intro H.
lin_top.
*)
lin_rever_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma with_forall :
  forall T0, forall a b : T0 -> Prop,
    ((forall x : T0, a x) [&] (forall x : T0, b x)) o-&-o forall x : T0, (a x [&] b x).
Proof.
(*
intros T0 a b.
lin_split_equiv.
intro W.
intro x.
lin_with.
lin_left_with_1 W.
specialize W with x.
lin_assumption.
lin_left_with_2 W.
specialize W with x.
lin_assumption.

intro F.
lin_with ;
intro x ;
specialize F with x.
lin_left_with_1 F.
lin_assumption.
lin_left_with_2 F.
lin_assumption.
*)
lin_rever_mall.
lin_left_with_1 H.
lin_apply0 H.
lin_left_with_2 H.
lin_apply0 H.
specialize H with x.
lin_left_with_1 H.
lin_assumption.
specialize H with x.
lin_left_with_2 H.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_exists :
  forall T0, forall a b : T0 -> Prop,
    (exists x : T0, a x) [+] (exists x : T0, b x) o-&-o exists x : T0, (a x [+] b x).
Proof.
(*
intros T0 a b.
lin_split_equiv.
intro P.
lin_left_plus P ;
destruct P ;
exists x.
lin_plus_1.
lin_assumption.
lin_plus_2.
lin_assumption.

intro E.
destruct E as [x P].
lin_left_plus P.
lin_plus_1.
exists x.
lin_assumption.
lin_plus_2.
exists x.
lin_assumption.
*)
lin_rever_mall.
exists x.
lin_auto_mall.
exists x.
lin_auto_mall.
lin_plus_1.
exists x.
lin_assumption.
lin_plus_2.
exists x.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma exists_with :
  forall T0, forall a b : T0 -> Prop,
    (exists x : T0, (a x [&] b x)) -> (exists x : T0, a x) [&] (exists x : T0, b x).
Proof.
(*
intros T0 a b E.
lin_with ;
destruct E as [x W] ;
exists x.
lin_left_with_1 W.
lin_assumption.
lin_left_with_2 W.
lin_assumption.
*)
lin_rever_mall ;
exists x ;
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_forall :
  forall T0, forall a b : T0 -> Prop,
    (forall x : T0, a x) [+] (forall x : T0, b x) -> forall x : T0, (a x [+] b x).
Proof.
intros T0 a b P x.
lin_left_plus P ;
specialize P with x.
lin_plus_1.
lin_assumption.
lin_plus_2.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.


(*== Additives and multiplicatives ==*)

Lemma top_absorb :
  forall a : Prop, (a -> [T]) o-&-o [T].
Proof.
(*
intro a.
lin_split_equiv.
intro I.
lin_top.
intros T Ha.
lin_top.
*)
lin_rever_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma lmap_with :
  forall a b c : Prop, (a -> b [&] c) o-&-o ((a -> b) [&] (a -> c)).
Proof.
(*
intros a b c.
lin_split_equiv.
intro I.
lin_with ;
intro Ha ;
lin_hyp_apply1 I in Ha.
lin_left_with_1 Ha.
lin_assumption.
lin_left_with_2 Ha.
lin_assumption.

intros W Ha.
lin_with.
lin_left_with_1 W.
lin_hyp_apply1 W in Ha.
lin_assumption.
lin_left_with_2 W.
lin_hyp_apply1 W in Ha.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma zero_absorb_r :
  forall a : Prop, (a /\ [0]) o-&-o [0].
Proof.
(*
intro a.
lin_split_equiv.
intro t.
lin_left_conj t.
lin_left_zero t1.
intro Z.
lin_left_zero Z.
*)
lin_rever_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma zero_absorb_l :
  forall a : Prop, ([0] /\ a) o-&-o [0].
Proof.
(*
intro a.
lin_split_equiv.
intro t.
lin_left_conj t.
lin_left_zero t0.
intro Z.
lin_left_zero Z.
*)
lin_rever_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma tens_plus :
  forall a b c : Prop, (a /\ (b [+] c)) o-&-o ((a /\ b) [+] (a /\ c)).
Proof.
intros a b c.
lin_split_equiv.
intro t.
lin_left_conj t.
lin_left_plus t1.
lin_plus_1.
lin_split with t0 ;
lin_assumption.
lin_plus_2.
lin_split with t0 ;
lin_assumption.
intro P.
lin_left_plus P ;
lin_left_conj P ;
lin_split with P0.
lin_assumption.
lin_plus_1.
lin_assumption.
lin_assumption.
lin_plus_2.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma tens_with :
  forall a b c : Prop, a /\ (b [&] c) -> (a /\ b) [&] (a /\ c).
Proof.
(*
intros a b c t.
lin_split_equiv.
lin_left_conj t.
lin_split with t0.
lin_assumption.
lin_left_with_1 t1.
lin_assumption.
lin_left_conj t.
lin_split with t0.
lin_assumption.
lin_left_with_2 t1.
lin_assumption.
*)
lin_rever_mall.
lin_left_with_1 H1.
lin_split with H0 ;
lin_assumption.
lin_left_with_2 H1.
lin_split with H0 ;
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_lmap :
  forall a b c : Prop, (a -> b) [+] (a -> c) -> a -> b [+] c.
Proof.
(*
intros a b c P.
lin_left_plus P ;
intro Ha ;
lin_hyp_apply1 P in Ha.
lin_plus_1.
lin_assumption.
lin_plus_2.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_lmap_with :
  forall a b c : Prop, (a -> c) [+] (b -> c) -> a [&] b -> c.
Proof.
(*
intros a b c P W.
lin_left_plus P.
lin_left_with_1 W.
lin_hyp_apply1 P in W.
lin_assumption.
lin_left_with_2 W.
lin_hyp_apply1 P in W.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma zero_lmap_left_top :
  forall a : Prop, ([0] -> a) o-&-o [T].
Proof.
(*
intros a.
lin_split_equiv.
intro I.
lin_top.
intros t Z.
lin_left_zero Z.
*)
lin_rever_mall.
(* llcheck
LL.
llcheck *)
Qed.

Lemma plus_lmap_left_with :
  forall a b c : Prop, ((a [+] b) -> c) o-&-o ((a -> c) [&] (b -> c)).
Proof.
(*
intros a b c.
lin_split_equiv.
intro I.
lin_with.
intro Ha.
lin_apply1 I.
lin_plus_1.
lin_assumption.
intro Hb.
lin_apply1 I.
lin_plus_2.
lin_assumption.

intros W P.
lin_left_plus P.
lin_left_with_1 W.
lin_apply1 W.
lin_assumption.
lin_left_with_2 W.
lin_apply1 W.
lin_assumption.
*)
lin_auto_mall.
(* llcheck
LL.
llcheck *)
Qed.


(*== Second order encodings ==*)

Lemma sndenc_negneg : forall a : Prop, a o-&-o forall X : Prop, (a -> X) -> X.
Proof with lin_rever_mall.
(*
intros a.
lin_split_equiv.
intros Ha X I.
lin_apply1 I.
lin_assumption.
intro I.
lin_apply1 I.
intro Ha.
lin_assumption.
*)
lin_rever_mall.
lin_auto_mall.
lin_apply1 H...
(* llcheck
LL.
llcheck *)
Qed.

Lemma sndenc_exists : forall T0, forall a : T0 -> Prop,
  (exists x : T0, a x) o-&-o forall X : Prop, (forall x : T0, (a x -> X)) -> X.
Proof with lin_rever_mall.
(*
intros T0 a.
lin_split_equiv.
intros He X I.
destruct He as [x].
specialize I with x.
lin_apply1 I.
lin_assumption.
intro I.
lin_apply1 I.
intros x Ha.
exists x.
lin_assumption.
*)
lin_rever_mall.
specialize H0 with x.
lin_apply1 H0...
lin_apply1 H...
exists x...
(* llcheck
LL.
llcheck *)
Qed.

Lemma sndenc_one : True o-&-o forall X : Prop, X -> X.
Proof with lin_rever_mall.
(*
lin_split_equiv.
intro O.
lin_left_one O.
intros X HX.
lin_assumption.
intros I.
lin_apply1 I.
lin_one.
*)
lin_rever_mall.
lin_apply1 H...
(* llcheck
LL.
llcheck *)
Qed.

Lemma sndenc_tens :
  forall a b : Prop, (a /\ b) o-&-o forall X : Prop, (a -> b -> X) -> X.
Proof with lin_rever_mall.
(*
intros a b.
lin_split_equiv.
intros t X i.
lin_left_conj t.
lin_apply2 i with t0;
assumption.

intro i.
specialize i with (a (x) b).
lin_apply1 i.
intros Ha Hb.
lin_split with Ha;
assumption.
*)
lin_rever_mall.
lin_apply2 H0 with H1...
lin_apply1 H...
lin_split with H...
(* llcheck
LL.
llcheck *)
Qed.

Lemma sndenc_zero : [0] o-&-o forall X : Prop, X.
Proof.
(*
lin_split_equiv.
intro O.
lin_left_zero O.
intros HX.
specialize HX with [0].
lin_assumption.
*)
lin_rever_mall.
lin_apply0 H.
(* llcheck
LL.
llcheck *)
Qed.

Local Close Scope mll_scope.
Local Close Scope all_scope.

