(*=== Tutorial v.0.2 for l2coq v.0.5.6 ===*)

Require Import l2coq. (* loading tactics and definitions *)

Local Open Scope mll_scope. (* notations for multiplicatives *)
Local Open Scope all_scope. (* notations for additives *)


(*= multiplicatives =*)

Goal 1. (* multiplicative unit *)
Proof.
lin_one. (* right introduction rule for 1 *)
(* llcheck
LL.
llcheck *)
Qed.

Goal forall a : Prop, a -o a.
  (* second order variables are of type Prop,
     "-o" is linear implication *)
Proof.
intro a. (* right introduction rule for quantifiers *)
intro Ha. (* right introduction rule for linear implication *)
lin_assumption. (* linear axiom rule *)
(* llcheck
LL.
llcheck *)
Qed.

Goal forall a b : Prop, a -o b -o a (x) b.
  (* "(x)" is tensor product *)
Proof.
intros a b.
intros Ha Hb.
lin_split with Ha. (* right introduction rule for tensor,
                      the context must be split,
                      the explicit context is given for left premisse,
                      the other hypotheses of sort Prop will be used for the right premisse,
                      hypotheses of sort Type (e.g. of type Prop) are copied on both sides *)
lin_assumption.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Goal forall a b : Prop, a (x) (1 (x) b) -o b (x) a.
Proof.
intros a b.
intros Ht.
lin_left_conj Ht. (* left introduction rule for tensor *)
lin_split with Ht1.
lin_left_conj Ht1.
lin_left_one Ht0. (* left introduction rule for 1 *)
lin_assumption.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Goal forall a b : Prop, (a -o b) -o a -o b.
Proof.
intros a b.
intros Hi Ha.
lin_left_imp Hi with. (* left introduction rule for linear implication,
                         the context has to be split, but no hypothesis is required for b *)
lin_assumption.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Goal forall a b : Prop, (a -o b) -o a -o b.
Proof.
intros a b.
intros Hi Ha.
lin_apply1 Hi. (* use an hypothesis of the shape _ -o G to prove the goal G,
                  index "1" corresponds to the arity (here 1) of the linear implication *)
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Goal forall a b : Prop, (a -o 1 -o b) -o a -o b.
Proof.
intros a b.
intros Hi Ha.
lin_apply2 Hi with Ha.
  (* use an hypothesis of the shape _ -o _ -o G to prove the goal G,
     index "2" corresponds to the arity (here 2) of the linear implication
     and thus a splitting of the context is required *)
lin_assumption.
lin_one.
(* llcheck
LL.
llcheck *)
Qed.

Goal forall a b : Prop, (a -o b) -o a -o b.
Proof.
intros a b.
intros Hi Ha.
lin_hyp_apply1 Hi in Ha.
  (* use an hypothesis of the shape H -o H' in hypothesis H to turn it into H',
     index "1" corresponds to the arity (here 1) of the linear implication *)
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Goal forall a b : Prop, (a -o 1 -o b) -o a -o (b -o b) -o b.
Proof.
intros a b.
intros Hi Ha Hb.
lin_hyp_apply2 Hi in Ha with.
  (* use an hypothesis of the shape H -o H' -o H'' in hypothesis H to turn it into H''
     and requires to prove H' as a new goal
     index "2" corresponds to the arity (here 2) of the linear implication
     and thus a splitting of the context is required
     the given context is used to prove H' *)
lin_apply1 Hb.
lin_assumption.
lin_one.
(* llcheck
LL.
llcheck *)
Qed.



(*= additives =*)

Goal forall a b : Prop, a -o b -o T. (* additive unit for with, terminal object *)
Proof.
intros a b.
intros Ha Hb.
lin_top. (* right introduction rule for top *)
(* llcheck
LL.
llcheck *)
Qed.

Goal forall a : Prop, a -o a (&) a. (* "(&)" is with, the cartesian product *)
Proof.
intro a.
intro Ha.
lin_with ; (* right introduction rule for with *)
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Goal forall a : Prop, a (+) a -o a (+) a. (* "(+)" is plus, the co-product *)
Proof.
intro a.
intro Hp.
lin_plus_1. (* right introduction rule 1 for plus *)
lin_left_plus Hp. (* left introduction rule for plus *)
lin_assumption.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Goal forall a b : Prop, a (&) 0 -o b. (* 0 is the additive unit for plus, initial object *)
Proof.
intros a b.
intro Hw.
lin_left_with_2 Hw. (* left introduction rule 2 for with *)
lin_left_zero Hw. (* left introduction rule for 0 *)
(* llcheck
LL.
llcheck *)
Qed.



(*= exponentials =*)

Goal !1 -o !1. (* ! is the exponential modality *)
Proof.
intro Ho.
lin_prom. (* promotion rule *)
lin_contr Ho. (* left contraction rule on ! formula *)
lin_weak Ho0. (* left weakening rule on ! formula *)
lin_der Ho. (* left dereliction rule *)
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.


(*= quantifiers =*)

Goal forall a : Prop, (forall b : Prop, b) -o a. (* second order quantification is done in Prop *)
Proof.
intro a. (* right introduction rule for forall *)
intro Hf.
specialize (Hf a). (* left introduction rule for forall *)
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Parameter ind : Set. (* individuals in Set for first order quantification *)
(* a first order signature over ind *)
Parameter c : ind. (* a constant symbol *)
Parameter f : ind -> ind. (* a unary symbol function *)
Parameter g : ind -> ind -> ind. (* a binary symbol function *)

Goal
  forall a : ind -> Prop, (* quantification over predicates on individuals *)
  forall x : ind, exists y : ind, (* first order quantifications *)
   a (f x) -o a y.
Proof.
intro a.
intro x. (* right introduction rule for forall *)
exists (f x). (* right introduction rule for exists *)
intro Ha.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Goal forall a : ind -> Prop, (forall x : ind, a x) -o a (g c c).
Proof.
intro a.
intro Hf.
specialize (Hf (g c c)). (* left introduction rule for forall *)
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.

Goal
  forall r : Prop, forall a : ind -> Prop,
  !((exists x : ind, forall y : ind, !(a x -o r) -> (!(a y) -o r)) -o r) -o r.
Proof.
intros r a.
intros Ho.
lin_contr Ho.
lin_der Ho0.
lin_apply1 Ho0.
exists c.
intro y.
intros Ha1 Ha2.
lin_der Ho.
lin_apply1 Ho.
exists y.
intro y0.
intros Ha3 Ha4.
lin_der Ha3.
lin_apply1 Ha3.
lin_der Ha2.
lin_weak Ha1.
lin_weak Ha4.
lin_assumption.
(* llcheck
LL.
llcheck *)
Qed.


(*= automation =*)

Require Import l2coq_auto.

Goal forall a : Prop, !a -o !(a (&) a).
Proof.
lin_auto. (* very elementary automatic tactic *)
(* llcheck
LL.
llcheck *)
Qed.

Goal forall a b : Prop, (!!a) (x) !b -o !(a (&) b).
intros.
lin_rever. (* apply all possible reversible rules *)
lin_der H0.
lin_easy. (* try axioms rules: one, top, axiom (with allowed !_ context) *)
(* llcheck
LL.
llcheck *)
Qed.



